import matplotlib.pyplot as plt
from read_cifar import *
import numpy as np

# Sigmoid activation function
def sigmoid(z):
    return 1 / (1 + np.exp(-z))

# Function to perform one gradient descent step with MSE loss
def learn_once_mse(w1, b1, w2, b2, data, labels, learning_rate):

    # Set input data, input dimension, hidden neurons, and output dimension
    m = data.shape[0]

    a0 = data
    z1 = np.matmul(a0, w1) + b1
    a1 = sigmoid(z1)
    z2 = np.matmul(a1, w2) + b2
    a2 = sigmoid(z2)

    # Compute Mean Squared Error (MSE) loss
    loss = np.mean(np.square(a2 - labels))

    # Calculate gradients
    d_a2 = 2 * (a2 - labels) / m
    d_z2 = d_a2 * a2 * (1 - a2)
    d_w2 = np.matmul(a1.T, d_z2)
    d_b2 = np.sum(d_z2, axis=0)

    d_a1 = np.matmul(d_z2, w2.T)
    d_z1 = d_a1 * a1 * (1 - a1)
    d_w1 = np.matmul(a0.T, d_z1)
    d_b1 = np.sum(d_z1, axis=0)

    # Update weights and biases
    w1 -= learning_rate * d_w1
    b1 -= learning_rate * d_b1
    w2 -= learning_rate * d_w2
    b2 -= learning_rate * d_b2

    return w1, b1, w2, b2, loss

'''
Function to code a vector to one-hot encoding
This function is general for any possible array, for a specific problem is better to use the number of possible classes to match 
the dimension in the further computation, in case of not all the possible classes are present in the 'labels' array.
For my algorithm it works well because I use the full training data in one step, so all the classes are surely present.
'''
def one_hot(labels):

    #num_classes = 10
    #one_hot_matrix = np.zeros((labels.shape[0], num_classes))

    one_hot_matrix = np.zeros((labels.shape[0], max(labels)+1))
    for i in range(labels.shape[0]):
        one_hot_matrix[i, labels[i]] = 1
    return one_hot_matrix

# Function to perform one gradient descent step with Binary cross-entropy loss
def learn_once_cross_entropy(w1, b1, w2, b2, data, labels, learning_rate):
    m = data.shape[0]  # Batch size

    a0 = data
    z1 = np.matmul(a0, w1) + b1
    a1 = sigmoid(z1)
    z2 = np.matmul(a1, w2) + b2
    a2 = sigmoid(z2)

    #Compute the Binary Cross-Entropy loss
    loss = -np.sum(labels * np.log(a2) + (1 - labels) * np.log(1 - a2)) / m

    # To update the weights and biases, we need to calculate the gradients of the loss with respect to each parameter
    d_a2 = (a2 - labels) / m
    d_z2 = d_a2
    d_w2 = np.matmul(a1.T, d_z2)
    d_b2 = np.sum(d_z2, axis=0, keepdims=True)

    d_a1 = np.matmul(d_z2, w2.T)
    d_z1 = d_a1 * a1 * (1 - a1)
    d_w1 = np.matmul(a0.T, d_z1)
    d_b1 = np.sum(d_z1, axis=0, keepdims=True)

    # Update weights and biases, these serve to reduce the loss during training. As is demonstrated in the plot 'loss.png'
    w1 -= learning_rate * d_w1
    b1 -= learning_rate * d_b1
    w2 -= learning_rate * d_w2
    b2 -= learning_rate * d_b2

    return w1, b1, w2, b2, loss

# Function to train an MLP for a specified number of epochs
def train_mlp(w1, b1, w2, b2, data_train, labels_train, learning_rate, num_epochs):

    train_accuracies = []
    for epoch in range(num_epochs):
        print('EPOCH ' + str(epoch + 1))
        labels_coded = one_hot(labels_train)
        w1, b1, w2, b2, loss = learn_once_cross_entropy(w1, b1, w2, b2, data_train, labels_coded, learning_rate)

        a0 = data_train
        z1 = np.matmul(a0, w1) + b1
        a1 = sigmoid(z1)
        z2 = np.matmul(a1, w2) + b2
        a2 = sigmoid(z2)
        predictions = np.argmax(a2, axis=1)

        # Compute the training accuracy for this epoch
        accuracy = np.mean(predictions == labels_train)
        train_accuracies.append(accuracy)
        print('Accuracy : ' + str(round(accuracy, 4)))
        print('Loss : '+ str(round(loss, 4)) + '\n')

    return w1, b1, w2, b2, train_accuracies

# Function to test the MLP on a test set
def test_mlp(w1, b1, w2, b2, data_test, labels_test):

    a0 = data_test
    z1 = np.matmul(a0, w1) + b1
    a1 = sigmoid(z1)
    z2 = np.matmul(a1, w2) + b2
    a2 = sigmoid(z2)

    predictions = np.argmax(a2, axis=1)
    test_accuracy = np.mean(predictions == labels_test)

    return test_accuracy

# Function to run the entire MLP training and testing process
def run_mlp_training(data_train, labels_train, data_test, labels_test, d_h, learning_rate, num_epochs):

    d_in = data_train.shape[1]
    num_classes = len(np.unique(labels_train))

    # Initialize the first and second layer weights and biases
    w1 = 2 * np.random.rand(d_in, d_h) - 1
    b1 = np.zeros((1, d_h))
    w2 = 2 * np.random.rand(d_h, num_classes) - 1
    b2 = np.zeros((1, num_classes))

    # Train the MLP on the training data
    w1, b1, w2, b2, train_accuracies = train_mlp(w1, b1, w2, b2, data_train, labels_train, learning_rate, num_epochs)

    # Test the MLP on the testing data
    test_accuracy = test_mlp(w1, b1, w2, b2, data_test, labels_test)

    return train_accuracies, test_accuracy

def main():

    print('#START#')

    # Set hyperparameters
    split_factor = 0.9
    d_h = 64
    learning_rate = 0.1
    num_epochs = 100

    # Load CIFAR dataset and split the training data and the labels for the two phases(train and test)
    folder_path = 'data/cifar-10-batches-py'
    data, labels = read_cifar(folder_path)
    data_train, data_test, labels_train, labels_test = split_dataset(data, labels, split_factor)

    # Run MLP training and testing
    train_accuracies, test_accuracy= run_mlp_training(data_train, labels_train, data_test, labels_test, d_h, learning_rate,
                                                       num_epochs)
    # Test accuracy after 'num_epochs' epochs of training
    print('FINAL ACCURACY : ' + str(round(test_accuracy, 4)) + '\n')

    # Plot the training accuracy for each epoch
    x = range(1, num_epochs + 1)
    plt.plot(x, train_accuracies)
    plt.xlabel('Epoch')
    plt.ylabel('Training Accuracy')
    plt.title('Training accuracy evolution')
    plt.grid()
    plt.savefig('results/mlp2.png')
    plt.show()

if __name__ == "__main__":
    main()


